package org.freecompany.util.text;

import java.nio.*;

/**
 * This class provides a wrapper view of a buffer of bytes as a sequence
 * of US-ASCII characters.  This is appropriate for treating a buffer of bytes
 * like a buffer of characters when it is known that it contains only 7 bit
 * characters.
 */
public class ByteBufferWrapper implements CharSequence {

	protected final ByteBuffer buffer;

	public ByteBufferWrapper( ByteBuffer buffer) {
		this.buffer = buffer;
	}

	public ByteBuffer getBuffer() {
		return buffer;
	}

	/**
	 * Returns the character at the provided index in the backing byte buffer, respective of
	 * the current buffer position.
	 */
	public char charAt( final int index) {
		return ( char) buffer.get( buffer.position() + index);
	}

	/**
	 * Returns the perceived length of this character sequence, which in this case is the number
	 * of bytes between the position and limit of the backing buffer.
	 */
	public int length() {
		return buffer.limit() - buffer.position();
	}

	/**
	 * Creates a new wrapped slice of the buffer given position and limit values
	 * derived from the provided start and end indexes.  After slicing, the underlying
	 * buffer is reset to it's original position and limit, but as the temporary changes
	 * are not atomic care must be taken not to depend on the state of the backing buffer
	 * being consistent during this call.  In other words, the buffer is not safe for access
	 * by other threads while this method is executing.
	 */
	public CharSequence subSequence( final int start, final int end) {
		final int size = buffer.limit() - buffer.position();
		try {
			buffer.limit( buffer.position() + end).position( buffer.position() + start);
			return new ByteBufferWrapper( buffer.slice());
		} finally {
			buffer.position( buffer.position() - start).limit( buffer.position() + size);
		}
	}

	/**
	 * Creates a UTF-8 decoded string representation of this byte buffer.  Teh behavior is undefined
	 * if the backing buffer contains any non US-ASCII characters.  This method is implemented to build
	 * a string one character at a time and so may be prohibitively slow for large buffers.
	 */
	public String toString() {
		final StringBuilder builder = new StringBuilder();
		final int length = buffer.limit();
		for ( int x = buffer.position(); x < length; x++) builder.append(( char) buffer.get( x));
		return builder.toString();
	}
}
