package org.freecompany.util.text;

/**
 * Locates individual characters and sequences of characters within a block of characters.
 * This class utilizes the {@link CharSequence} interface for scan targets, allowing a wide
 * array of container objects to be searched.
 * <p/>
 * The scanner provides set of static methods to perform character sequence scanning relative
 * to either a character sequence or a state object as provided by the caller.
 * <p/>
 */
public class Scanner {

	/**
	 * Scans for the next occurance of the provided sequence starting at the index stored in the internal
	 * counter.  The internal counter will be updated to the end of the matched sequence or the next index
	 * after the last possible match of the sequence.  This is so that the target can be reset with a longer
	 * sequence and scanning will resume at the next possible index where the match can occur.
	 *
	 * @return the index of the next match, or -1 if no match was found.
	 */
	public final static int scan( final State state, final CharSequence sequence) {
		final CharSequence target = state.sequence();
		final int end = end( target, sequence);
		final int match = scan( target, sequence, state.position(), end);
		state.position( match == -1 ? end : match + sequence.length());
		return match;
	}

	/**
	 * Skips a sequence if it is present at the cursor.  This method will reset the cursor to the new position
	 * if a match is found or leave it untouched if not.
	 */
	public final static boolean skip( final State state, final CharSequence sequence) {
		final int match = skip( state.sequence(), sequence, state.position());
		if ( match != state.position()) {
			state.position( match);
			return true;
		}
		return false;
	}

	/**
	 * Convenience method that scans in order to determine if the token is the first set of characters in
	 * the target.
	 */
	public final static boolean startsWith( final CharSequence target, final CharSequence sequence) {
		return match( target, sequence, 0);
	}

	/**
	 * Convenience method that scans in order to determine if the token is the last set of characters in
	 * the target.
	 */
	public final static boolean endsWith( final CharSequence target, final CharSequence sequence) {
		return match( target, sequence, target.length() - sequence.length());
	}

	/**
	 * Convenience method that reverse scans an entire target.  This is equivalent to
	 * calling <code>scan</code> with the start and end arguments reversed.
	 */
	public final static int last( final CharSequence target, final CharSequence sequence) {
		return scan( target, sequence, target.length() - 1, 0);
	}

	/**
	 * Convenience method that scans between the start and the end of the target.  This is equivalent to
	 * calling <code>scan</code> with zero as the third argument and the length of the sequence subtracted
	 * from the length of the target as the fourth argument.
	 */
	public final static int scan( final CharSequence target, final CharSequence sequence) {
		return scan( target, sequence, 0);
	}

	/**
	 * Convenience method which scans between the provided index and the end of the target.  This is equivalent
	 * to calling the <code>scan</code> method with the length of the sequence subtracted from the length of the target
	 * as the fourth argument.
	 */
	public final static int scan( final CharSequence target, final CharSequence sequence, final int start) {
		return scan( target, sequence, start, end( target, sequence));
	}

	/**
	 * Scans the provided target sequence for the whole of the provided sequence occurring between the start and
	 * end indexes.  Reverse scans if the start is greater than the end.
	 *
	 * @return the index of the first character of the sequence in the target, or -1 if the sequence was not found.
	 */
	public final static int scan( final CharSequence target, final CharSequence sequence, int start, final int end) {
		if ( start < end) {
			for (; start < end; start++) if ( match( target, sequence, start)) return start;
		} else {
			for (; end <= start; start--) if ( match( target, sequence, start)) return start;
		}
		return -1;
	}

	/**
	 * Skips a sequence using the default index of zero, meaning this method only searches the beginning of a sequence
	 * for a particular match.
	 *
	 */
	public final static int skip( final CharSequence target, final CharSequence sequence) {
		return skip( target, sequence, 0);
	}

	/**
	 * Skips a sequence if and only if it is found at the given index.
	 *
	 * @return the new index after possibly skipping a sequence.  If the sequence was found then it returns the index
	 * after advancing past the match.  If not, the current index is returned.
	 */
	public final static int skip( final CharSequence target, final CharSequence sequence, int index) {
		return match( target, sequence, index) ? index + sequence.length() : index;
	}

	/**
	 * Convenience method that matches the portion of the sequence from the provided index to the end.  This is
	 * equivalent to calling the <code>match</code> method with the length of the sequence added to the start
	 * index as the fourth argument.
	 */
	public final static boolean match( final CharSequence target, final CharSequence sequence, final int start) {
		return match( target, sequence, start, start + sequence.length());
	}

	/**
	 * Matches the provided sequence against the target at the indicated index.  This method tests that the target
	 * contains as many of this exact sequences characters as fit between the start and end indexes.
	 *
	 * @return true if the target contains the sequence between the start and end positions, ignoring extra characters
	 * or false otherwise.
	 */
	public final static boolean match( final CharSequence target, final CharSequence sequence, int start, final int end) {
		final int base = start;
		for (; start < end; start++) if ( !Comparison.equals( target.charAt( start), sequence.charAt( start - base))) return false;
		return true;
	}

	protected final static int end( final CharSequence target, final CharSequence sequence) {
		return target.length() - sequence.length() + 1;
	}
}
