package org.freecompany.util.text;

import java.nio.ByteBuffer;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharacterCodingException;
import java.util.Iterator;

/**
 * Encompases the state of a text processor.  This includes a view of the
 * currently available input data, complete with automatic management functions,
 * as well as a position and mark variable.  The position is the position of the
 * last character processed, and the mark is a place holder for recalling the
 * start of the current input chunk.
 * <p/>
 * It should be noted that in general the performance of text operations on this
 * state object will be slightly higher when the entirety of the input sequence
 * is held within a single buffer.  In other words, when possible it is preferrable
 * to provide the complete input sequence with a single {@link #append( CharSequence sequence)}
 * invocation.
 * <p/>
 * By utilizing the {@link CharSequence} interface exclusively, this scanner can natively supports
 * {@link String}, {@link CharBuffer}, {@link StringBuilder} and {@link StringBuffer} objects as
 * input sequences, as well as specially wrapped {@link java.nio.ByteBuffer} instances.
 */
public class State implements CharSequence, Iterator< Character>, Iterable< Character> {

	protected CharSequence sequence = CharSequenceAggregate.empty();
	protected int position;
	protected int mark;

	public void clear() {
		this.sequence = CharSequenceAggregate.empty();
		this.position = 0;
		this.mark = 0;
	}

	/**
	 * Appends a new sequence to the end of the existing one, allowing the sequence to grow
	 * in place.  This is handled by wrapping the old and new sequences in an aggregating
	 * object.  If no sequence exists this sequence is set as the new sequence and no wrapping
	 * occurs.
	 */
	public void append( final CharSequence sequence) {
		if ( sequence == null) throw new NullPointerException();
		this.sequence = CharSequenceAggregate.add( this.sequence, sequence);
	}

	/**
	 * The buffer is assumed to be US-ASCII encoded text (with undefined behavior if
	 * this is not the case) and is simply wrapped to look like a character sequence.
	 * In particular, this method will not work if the buffer contains character data
	 * encoded in any other character set, including <code>UTF-8</code>.
	 * <p/>
	 * The buffer is used as is and must be preserved until this state object is no longer
	 * needed, and any sub-sequences or slices have been discarded.
	 */
	public void append( final ByteBuffer buffer) {
		append( new ByteBufferWrapper( buffer));
	}

	/**
	 * Appends a new buffer to the target using the provided decoder.  A new {@link CharBuffer}
	 * is allocated from the pool and the buffer is decoded into it in it's entirety.
	 *
	 * The caller is free to reuse or dispose of the passed buffer immediately.
	 */
	public void append( final ByteBuffer buffer, final CharsetDecoder decoder) throws CharacterCodingException {
		append( decoder.decode( buffer));
	}

	/**
	 * Performs a soft expunge which only removes the first sequences that have been exhausted.  This
	 * is the same as calling {@link #expunge( boolean force)} with the argument <code>false</code>.
	 */
	public CharSequence expunge() {
		return expunge( false);
	}

	/**
	 * Removes the next unneeded sequence from this parser, if one is available.  If these sequences
	 * wrap shared memory they cannot be reused until after all external references have been cleared.
	 * If the <code>force</code> flag is specified, then the first {@link CharSequence} will be
	 * removed regardless of the current position of this scanner.
	 *
	 * @return the next <@see ByteBuffer/> if one is available or null if all are
	 * still in use or if the parser contains non <@see ByteBuffer/> backed <@see CharSequence/>s.
	 */
	public CharSequence expunge( final boolean force) {
		final CharSequence root;
		if ( sequence instanceof CharSequenceAggregate) root = (( CharSequenceAggregate) sequence).root();
		else root = sequence;

		if ( root != CharSequenceAggregate.NULL && ( force || position >= root.length())) {
			sequence = CharSequenceAggregate.remove( sequence, root);
			position = Math.max( 0, position - root.length());
			mark = Math.max( 0, mark - root.length());
			return root;
		}

		return null;
	}

	public CharSequence sequence() { return sequence; }

	public CharSequence subSequence( final int from, final int to) { return sequence.subSequence( from, to); }
	public char charAt( final int index) { return sequence.charAt( index); }
	public int length() { return sequence.length(); }

	public Iterator< Character> iterator() {
		return this;
	}

	/**
	 * Slices the most recently processed logical characters sequence by capturing and returning
	 * the sequence of characters between the mark and the current position and then marking the
	 * new position.  This prepares the state for the next processing operation.
	 */
	public CharSequence slice() { 
		try {
			return sequence.subSequence( mark, position);
		} finally {
			mark();
		}
	}

	/**
	 * Returns the character immediately preceeding the current character in this sate object.
	 */
	public final char previous() {
		return sequence.charAt( position - 1);
	}

	/**
	 * Examines the next character without advancing the position.
	 */
	public final char peek() {
		return sequence.charAt( position);
	}

	/**
	 * Examines the character at the position plus the argument without
	 * advancing the position.
	 */
	public final char peekPlus( final int advance) {
		return sequence.charAt( position + advance);
	}

	public final Character next() {
		return sequence.charAt( position++);
	}

	public final int remaining() {
		return sequence.length() - position;
	}

	public final boolean hasNext() {
		return sequence.length() > position;
	}

	public final CharSequence rest() {
		position( sequence.length());
		return slice();
	}

	public final void position( int position) {
		this.position = position;
	}
	
	public final int position() {
		return position;
	}

	public boolean marked() {
		return mark != 0 && mark != position;
	}

	public final void mark() {
		mark = position;
	}

	public final int getMark() { return mark; }

	public final void reset() {
		position = mark;
	}

	public void remove() { throw new UnsupportedOperationException(); }
}
