package org.freecompany.util.text;

/**
 * Provides methods to tokenize character sequence given either absolute
 * indexes or sequence for which to scan.
 *
 * TODO: Make the whitespace boolean a ctor argument and refactor out a trim
 * method.
 */
public class Tokenizer extends Scanner {

	protected final static CharSequence sub( final CharSequence target, int start, int end, final boolean trim) {
		if ( trim) {
			while ( start < end - 1 && Character.isWhitespace( target.charAt( start))) start++;
			while ( end > start && Character.isWhitespace( target.charAt( end - 1))) end--;
		}
		return target.subSequence( start, end);
	}

	/**
	 * Retrieves the characters preceeding the provided sequence, starting at the current position.  The
	 * position is reset to the character following the match only if the match is found.
	 *
	 * @return the characters preceeding the sequence, or null if the sequence is not found.
	 * Retrieves the next token delimited by the provided sequence.
	 */
	public final static CharSequence before( final TokenizerState state, final CharSequence sequence) {
		state.mark();
		final int start = state.position();
		final int end = scan( state, sequence);
		if ( end == -1) {
			state.reset();
			return null;
		}

		return sub( state.sequence(), start, end, state.trim());
	}

	/**
	 * Retrieves the remaining characters after the provided sequence, starting at the current position.  The
	 * position is moved to the end of the target only if a match is found.
	 *
	 * @return the characters remaining after the sequence, or null if the sequence is not found.
	 */
	public final static CharSequence after( final TokenizerState state, final CharSequence sequence) {
		state.mark();
		if ( scan( state, sequence) == -1) {
			state.reset();
			return null;
		}

		final CharSequence target = state.sequence();
		try {
			return sub( target, state.position(), target.length(), state.trim());
		} finally {
			state.position( target.length());
		}
	}

	/**
	 * Retrieves the next token that falls between the two provided sequences, starting at the current position.  The
	 * position is moved to the character following the match only if the match is found.
	 *
	 * @return the characters between the two sequences, or null if either sequence is not found.
	 */
	public final static CharSequence between( final TokenizerState state, final CharSequence one, final CharSequence two) {
		state.mark();
		if ( scan( state, one) == -1) {
			state.reset();
			return null;
		}
		
		int start = state.position();
		int end = scan( state, two);
		if ( end == -1) {
			state.reset();
			return null;
		}

		return sub( state.sequence(), start, end, state.trim());
	}

	/**
	 * Retrieves the remaining characters from this tokenizer.
	 */
	public final static CharSequence rest( final TokenizerState state) {
		int start = state.position();
		int end = state.sequence().length();
		state.position( end);
		return sub( state.sequence(), start, end, state.trim());
	}

	/**
	 * Retrieves the characters preceeding the provided sequence.
	 *
	 * @return the characters preceeding the sequence, or null if the sequence is not found.
	 */
	public final static CharSequence before( final CharSequence target, final CharSequence sequence) {
		return before( target, sequence, false);
	}

	/**
	 * Retrieves the characters preceeding the provided sequence, optionally trimming the resulting sequence.
	 *
	 * @return the characters preceeding the sequence, or null if the sequence is not found.
	 */
	public final static CharSequence before( final CharSequence target, final CharSequence sequence, final boolean trim) {
		final int start = scan( target, sequence);
		return start == -1 ? null : sub( target, 0, start, trim);
	}

	/**
	 * Convenience method for returning the untrimmed portion of a sequence up to the provided index.  Equivalent to
	 * calling <code>before</code> with false as the last parameter.
	 */
	public final static CharSequence before( final CharSequence target, final int start) {
		return before( target, start, false);
	}

	/**
	 * Retrieves the characters preceding the provided index, optionally trimming the results.
	 *
	 * @return the characters remaining after the sequence, or null if the sequence is not found.
	 */
	public final static CharSequence before( final CharSequence target, final int start, final boolean trim) {
		return ( start < 0 || start >= target.length()) ? null : sub( target, 0, start, trim);
	}

	/**
	 * Retrieves the remaining characters after the provided sequence.
	 *
	 * @return the characters remaining after the sequence, or null if the sequence is not found.
	 */
	public final static CharSequence after( final CharSequence target, final CharSequence sequence) {
		return after( target, sequence, false);
	}

	/**
	 * Retrieves the remaining characters after the provided index.
	 *
	 * @return the characters remaining after the sequence, or null if the sequence is not found.
	 */
	public final static CharSequence after( final CharSequence target, final int index) {
		return after( target, index, false);
	}

	/**
	 * Retrieves the remaining characters after the provided sequence, optionally trimming the results.
	 *
	 * @return the characters remaining after the sequence, or null if the sequence is not found.
	 */
	public final static CharSequence after( final CharSequence target, final CharSequence sequence, final boolean trim) {
		final int start = scan( target, sequence);
		return start == -1 ? null : sub( target, start + sequence.length(), target.length(), trim);
	}

	/**
	 * Retrieves the remaining characters after the provided index, optionally trimming the results.
	 *
	 * @return the characters remaining after the sequence, or null if the sequence is not found.
	 */
	public final static CharSequence after( final CharSequence target, final int start, final boolean trim) {
		return ( start < 0 || start >= target.length()) ? null : sub( target, start + 1, target.length(), trim);
	}

	/**
	 * Retrieves the next token that falls between the two provided sequences.
	 *
	 * @return the characters between the two sequences, or null if either sequence is not found.
	 */
	public final static CharSequence between( final CharSequence target, final CharSequence one, final CharSequence two) {
		return between( target, one, two, false);
	}

	/**
	 * Retrieves the next token that falls between the two provided sequences, optionally trimming the results.
	 *
	 * @return the characters between the two sequences, or null if either sequence is not found.
	 */
	public final static CharSequence between( final CharSequence target, final CharSequence one, final CharSequence two, final boolean trim) {
		final int start = scan( target, one);
		if ( start == -1) return null;
		final int end = scan( target, two, start + 1);
		return end == -1 ? null : sub( target, start + one.length(), end, trim);
	}
}
